import Foundation
import Turf
import Mapbox

let temporaryDirectoryURL = URL(fileURLWithPath: NSTemporaryDirectory(), isDirectory: true)

struct Style: Codable, Equatable, Hashable {
  var version = 8
  
  typealias Source = AnyCodable
  typealias Layer = AnyCodable
  
  var sources: [String: Source]
  var layers: [Layer]
  
  typealias Sprite = String
  typealias Glyphs = String
  typealias Terrain = AnyCodable
  
  var sprite: Sprite? = nil
  var glyphs: Glyphs? = nil
  var terrain: Terrain? = nil
  
  var interfacedLayers: [InterfacedLayer] {
    layers.compactMap(InterfacedLayer.create)
  }
  
  var interfacedSources: [InterfacedSource] {
    sources.compactMap(InterfacedSource.create)
  }

  var zoomLevelsCovered: (min: Double, max: Double) {
    (
      min: interfacedSources.map({$0.minZoom ?? 0}).min() ?? 0, // is this what we want? need to test
      max: interfacedSources.map({$0.maxZoom ?? 22}).max() ?? 22
    )
  }
  
  var bounds: BoundsInfo {
    var superbound: MGLCoordinateBounds?
    var allBounds: [MGLCoordinateBounds] = []
    
    for source in interfacedSources {
      guard let _bounds = source.bounds else {
        // if a source has no bounds, then assume its worldwide - therefore short circuit, discarding bounds
        return BoundsInfo(individual: [], superbound: nil)
      }
      
      superbound = superbound?.extend(with: _bounds) ?? _bounds
      allBounds.append(_bounds)
    }
    
    return BoundsInfo(individual: allBounds, superbound: superbound)
  }
  
  var url: URL? {
    do {
      let encoder = JSONEncoder()
      
      let data = try encoder.encode(self)
      
      let temporaryFilename = UUID().uuidString
      let temporaryFileURL = temporaryDirectoryURL.appendingPathComponent(temporaryFilename)
      
      try data.write(to: temporaryFileURL, options: .atomic)
      
      return temporaryFileURL
    }
    catch {
      return nil
    }
  }
  
  var supportsOpacity: Bool {
    interfacedLayers.contains(where: {$0.capabilities.contains(.opacity)})
  }
  
  var opacity: Double {
    interfacedLayers.compactMap({
      switch $0.type {
        case "background", "raster": return $0.opacity
        default: return nil
      }
    }).max() ?? 1
  }
  
  var supportsColour: Bool {
    interfacedLayers.contains(where: {$0.capabilities.contains(.colour)})
  }
  
  var colour: UIColor? {
    interfacedLayers.first(where: {$0.colour != nil})?.colour
  }
  
  var hasData: Bool {
    interfacedSources.contains(where: {$0.capabilities.contains(.data)})
  }
  
  var hasNetworkDependencies: Bool {
    return (
      sprite != nil ||
      glyphs != nil ||
      interfacedSources.contains(where: {!$0.capabilities.contains(.data)})
    )
  }
  
  var interfacedSourcesContainingData: [InterfacedSource] {
    interfacedSources.filter({$0.capabilities.contains(.data)})
  }
  
  var identifiersOfLayersWithDataContainingSources: [String] {
    interfacedLayers
      .compactMap({layer in
        let hasDataContainingSource = interfacedSourcesContainingData.contains(where: {source in source.id == layer.source})
        
        return hasDataContainingSource ? layer.id : nil
      })
  }
  
  var geoJSONIdentifiableFeatures: [String: Feature] {
    let geoJSONFeaturesTuple: [(String, Feature)] = interfacedSourcesContainingData.flatMap({(source) -> [(String, Feature)] in
      let geoJSONData = source.geoJSONData

      guard let geoJSON = GeoJSONObject.from(anyCodable: geoJSONData)?.geoJSONObject
      else {return []}
      
//      guard let data = try? JSONEncoder().encode(geoJSONData),
//            let geoJSON = try? JSONDecoder().decode(GeoJSONObject.self, from: data)
//      else {return []}
    
      
      return geoJSON.flattenFeatures().compactMap(({feature -> (String, Feature)? in
        switch feature.identifier {
          case .string(let string): return (string, feature)
          default: return nil
        }
      }))
    })
    
    return Dictionary(uniqueKeysWithValues: geoJSONFeaturesTuple)
  }
  
  struct BoundsInfo {
    let individual: [MGLCoordinateBounds]
    let superbound: MGLCoordinateBounds?
  }
}
